
{******************************************}
{                                          }
{             FastReport VCL               }
{         Helper classes for Exports       }
{                                          }
{         Copyright (c) 1998-2021          }
{            by Fast Reports Inc.          }
{                                          }
{******************************************}

unit frxExportHelpers;

interface

{$I frx.inc}

uses
{$IFNDEF Linux}
  Windows,
{$ELSE}
  LCLType, LCLIntf, LCLProc,
{$ENDIF}
  Classes, Graphics, Controls,
  frxClass, frxExportBaseDialog,
{$IFDEF DELPHI16}
  System.UITypes,
{$ENDIF}
  Variants,      //for Outline
  frxCrypto,    // for hashing pictures and CSS styles
  frxStorage,   // for TCachedStream
  frxPictureGraphics, frxVectorCanvas, frxUtils, frxAnaliticGeometry, frxTrueTypeCollection, math,
{$IFDEF FPC}LazHelper, {$ENDIF}
  frxTrueTypeFont, frxPlatformServices, frxCSSStyle, frxBaseGraphicsTypes;

const
  DefaultPrec = 3;

type
   { For Outline }
  TfrxCustomOutlineNode = class
  public
    Number: Integer;
    Dest: Integer; // Index to a page referred to by this outline node
    Top: Integer; // Position on the referred to page
    CountTree: Integer; // Number of all descendant nodes
    Count: Integer; // Number of all first-level descendants
    Title: string;

    First: TfrxCustomOutlineNode; // The first first-level descendant
    Last: TfrxCustomOutlineNode; // The last first-level descendant
    Next: TfrxCustomOutlineNode; // The next neighbouring node
    Prev: TfrxCustomOutlineNode; // The previous neighbouring node
    Parent: TfrxCustomOutlineNode; // The parent node of this node

    constructor Create;
    destructor Destroy; override;
  end;

  { For PDF }

  TfrxPDFOutlineNode = class(TfrxCustomOutlineNode);

  { For HTML }

  TfrxHTMLExportGetNavTemplate = procedure(const ReportName: String;
    Multipage: Boolean; PicsInSameFolder: Boolean; Prefix: String;
    TotalPages: Integer; var Template: String) of object;

  TfrxHTMLExportGetMainTemplate = procedure(const Title: String;
    const FrameFolder: String;
    Multipage: Boolean; Navigator: Boolean; var Template: String) of object;

  TfrxHTMLExportGetToolbarTemplate = procedure(CurrentPage: Integer; TotalPages: Integer;  Multipage: Boolean; Naviagtor: Boolean; var Template: String) of object;

  TfrxHTMLOutlineNode = class(TfrxCustomOutlineNode);

  { Represents a CSS style }

  { Represents a CSS (Cascading Style Sheet) with a list of CSS styles }

  { Saves pictures and ensures that there will not be identical copies saved }

  TfrxPictureInfo = record
    Extension: string;
    Mimetype: string;
  end;

  { Generalised picture }

  TfrxPictureFormat = (pfPNG, pfEMF, pfBMP, pfJPG, pfBMP32);

  TfrxPicture = class
  private
    FFormat: TfrxPictureFormat;
    FGraphic: TGraphic;
    FVectorGraphic: TGraphic;
    FGHelper: TfrxCustomGraphicFormatClass;
    FGVectorHelper: TfrxCustomGraphicFormatClass;
    FCanvasHelper: TfrxGraphicCanvasHelper;
    FPixelFormat: TPixelFormat;
    FIsAlpha: Boolean;
    function GetCanvas: TCanvas; // for TJPEGImage that doesn't provide a canvas
  public
    constructor Create(Format: TfrxPictureFormat; Width, Height: Integer; Transparent: Boolean; IsVectorSource: Boolean; IsAlpha: Boolean);
    destructor Destroy; override;

    function Release: TGraphic;
    procedure SetTransparentColor(TransparentColor: TColor);
    procedure FillColor(Color: TColor; clNoneReplacement: TColor = clWhite);

    property Canvas: TCanvas read GetCanvas;
  end;

  TfrxExportHandler = function(Obj: TfrxView): Boolean of object;

  TTextFragment = class
  private
    FFormatted: boolean;
    FText: string;
  public
    constructor Create(AFormatted: boolean);
    procedure Add(const s: string); overload;
    procedure Add(const Fmt: string; const Args: array of const); overload;

    property Text: string read FText;
  end;

  TAnsiMemoryStream = class(TMemoryStream)
  private
    procedure PutsRaw(const s: AnsiString);
    procedure PutsA(const s: AnsiString);
  protected
    FFormatted: Boolean;
  public
    constructor Create(AFormatted: boolean = False);
    procedure Puts(const s: string); overload;
    procedure Puts(const Fmt: string; const Args: array of const); overload;

    function AsAnsiString: AnsiString;
  end;

  TExportHTMLDivSVGParent = class (TfrxBaseDialogExportFilter)
  private
    FServer: Boolean;
    FMultiPage: Boolean;
    FFormatted: Boolean;
    FPicFormat: TfrxPictureFormat;
    FUnifiedPictures: Boolean;
    FNavigation: Boolean;
    FEmbeddedPictures: Boolean;
    FEmbeddedCSS: Boolean;
    FOutline: Boolean;
    FWorkDir: string;
    FPrefix: string;
    procedure SetPicFormat(Fmt: TfrxPictureFormat);
    procedure SetMultiPage(const Value: Boolean);
    function SavePicInPath(Pic: TGraphic; Filter: TfrxCustomIOTransport; PNum: Integer): String;
  protected
    FFilterStream: TStream;
    FCurrentPage: Integer; // 1-based index of the current report page
    FCSS: TfrxCSSList; // stylesheet for all pages
    FHandlers: array of TfrxExportHandler;
    FCurrentFile: TStream;

    function EnableCalculateHash: Boolean; override;

    procedure AttachHandler(Handler: TfrxExportHandler);
    procedure RunExportsChain(Obj: TfrxView); virtual;
    function GetCSSFileName: string;
    function GetCSSFilePath: string;
    procedure SaveCSS(const FileName: string);
    procedure CreateCSS; virtual; abstract;
    function IsCanSavePicture(Pic: TGraphic): Boolean;
    function GetPicPath(Pic: TGraphic; PNum: Integer): String;
    function SavePicture(Pic: TGraphic; PNum: Integer = -1): String;
    procedure FreeStream;

    { Writes a string to the current file }
    procedure PutsRaw(const s: AnsiString);
    procedure PutsA(const s: AnsiString);
    procedure Puts(const s: string); overload;
    procedure Puts(const Fmt: string; const Args: array of const); overload;

    { Registers a CSS style in the internal CSS table and returns a selector
      that can be used in the "class" attribute of tags. }
    function LockStyle(Style: TfrxCSSStyle): string;

    function ExportViaVector(Obj: TfrxView): AnsiString;
    procedure Vector_ExtTextOut(Obj: TfrxView; AMS: TAnsiMemoryStream;
      Vector: TVector_ExtTextOut; const Shift: TPoint);
  public
    constructor Create(AOwner: TComponent); override;
    procedure ExportObject(Obj: TfrxComponent); override;
    function Start: Boolean; override;
    procedure StartPage(Page: TfrxReportPage; Index: Integer); override;

    property Server: Boolean read FServer write FServer;
  published
    property OverwritePrompt;
    property OpenAfterExport;
    { Exports each report page to a separate page }
    property MultiPage: Boolean read FMultiPage write SetMultiPage;
    { Makes HTML sources formatted (and sligtly bigger) }
    property Formatted: Boolean read FFormatted write FFormatted;
    { Format for pictures representing report objects that are not saved natively,
      like RichText objects. }
    property PictureFormat: TfrxPictureFormat read FPicFormat write SetPicFormat;
    { Converts all pictures to PictureFormat: if there's a BMP picture in a report
      and PictureFormat is PNG, then this BMP will be saved as a PNG. }
    property UnifiedPictures: Boolean read FUnifiedPictures write FUnifiedPictures;
    { Creates navigation controls for Multipage mode }
    property Navigation: Boolean read FNavigation write FNavigation;
    { Embeds pictures }
    property EmbeddedPictures: Boolean read FEmbeddedPictures write FEmbeddedPictures;

    { Embeds CSS stylesheet }
    property EmbeddedCSS: Boolean read FEmbeddedCSS write FEmbeddedCSS;
    { Embeds HTML Outline }
    property Outline: Boolean read FOutline write FOutline;
  end;

  TRotation2D = class
  private
    FCenter: TfrxPoint;
    FMatrix: String;
  protected
    Sinus, Cosinus: Extended;
    C1, C2: Extended;

  public
    procedure Init(Radian: Extended; Center: TfrxPoint; Precision: Integer = DefaultPrec);

    function Turn2Str(DP: TfrxPoint): string;
    function Turn(DP: TfrxPoint): TfrxPoint;

    property Matrix: String read FMatrix;
  end;

  SCRIPT_CACHE = Pointer;
  PScriptCache = ^SCRIPT_CACHE;

  SCRIPT_ANALYSIS = record
    fFlags: Word;
    s: Word
  end;
  PScriptAnalysis = ^SCRIPT_ANALYSIS;

  TfrxPDFRun = class
  public
    analysis: SCRIPT_ANALYSIS;
    text: WideString;
    constructor Create(t: WideString; a: SCRIPT_ANALYSIS);
  end;

  TRemapedString = record
    Data: WideString;
    Width: Integer;
    SpacesCount: Integer;
    IsValidCharWidth: Boolean;
    CharWidth: TIntegerDinArray;
    IsHasLigatures: Boolean;
    IsSpace: array of Boolean;
  end;

  TfrxExportFont = class
  private
    FUSCache: PScriptCache;
    FForceAnsi: Boolean;
    FSameCharacterWidth: Boolean;

    function GetGlyphs(hdc: HDC; run: TfrxPDFRun; glyphs: PWord; widths: PInteger; maxGlyphs: integer; rtl: boolean; IsIndexes: Boolean = false): Integer;
    function Itemize(s: WideString; rtl: boolean; maxItems: Integer): TList;
    function Layout(runs: TList; rtl: boolean): TList;
    function GetGlyphIndices(hdc: HDC; text: WideString; glyphs: PWord; widths: PInteger; rtl: boolean; IsIndexes: Boolean = false): integer;
  protected
    GlobalTempBitmap: TBitmap;
    TrueTypeTables: TrueTypeCollection;
    ttf:  TrueTypeFont;
    PackFont: Boolean;
    PDFdpi_divider: double;
    FDpiFX: double;
    FIsLigatureless: Boolean;

    procedure GetFontFile;
    function RemapString(str: WideString; rtl: Boolean; IsIndexes: Boolean = false): TRemapedString;
  public
    SourceFont: TFont;
    Widths: TList;
    UsedAlphabet: TList;
    UsedAlphabetUnicode: TList;
    TextMetric: {$IFDEF FPC}{$IFDEF Linux}OUTLINETEXTMETRIC{$ELSE}LPOUTLINETEXTMETRICA{$ENDIF}{$ELSE}^OUTLINETEXTMETRICA{$ENDIF};
    {$IFDEF Linux}
    FontData:  TMemoryStream;
    function GetFontDataSize(): Longint; Virtual; Abstract;
    property FontDataSize: Longint read GetFontDataSize;
    {$ELSE}
    FontData: PAnsiChar;
    FontDataSize: Longint;
    {$ENDIF}
    constructor Create(Font: TFont);
    destructor Destroy; override;
    function SoftRemapString(str: WideString; rtl: Boolean; IsIndexes: Boolean = false): TRemapedString;

    property ForceAnsi: Boolean write FForceAnsi;
    property SameCharacterWidth: Boolean write FSameCharacterWidth;
  end;

  TOnProcessHyperLink = function(HL: TfrxHyperlink; ParentDetailURL: String): String;

{ HTML Outline }
const  outlineCSS =
    '.outlineCSS,   '+
    '.outlineCSS ul,  '+
    '.outlineCSS li { '+
    '  margin: 0;  '+
    '  padding: 0;  '+
    '  line-height: 1; '+
    '  list-style: none; '+
    '}  '+
    '.outlineCSS ul { '+
    '  margin: 0 0 0 .5em; '+
    '}  '+
    '.outlineCSS > li:not(:only-child), '+
    '.outlineCSS li li { '+
    '  position: relative;'+
    '  padding: .2em 0 0 1.2em;'+
    '}  '+
    '.outlineCSS li:not(:last-child) {'+
    '  border-left: 1px solid #ccc; '+
    '}  '+
    '.outlineCSS li li:before,'+
    '.outlineCSS > li:not(:only-child):before {'+
    '  content: ""; '+
    '  position: absolute;  '+
    '  top: 0; '+
    '  left: 0; '+
    '  width: 1.1em; '+
    '  height: .7em;  '+
    '  border-bottom: 1px solid #ccc; '+
    '} '+
    '.outlineCSS li:last-child:before { '+
    '  width: calc(1.1em - 1px); '+
    '  border-left: 1px solid #ccc; '+
    '} '+
    '.outlineCSS .drop { '+
    '  position: absolute; '+
    '  left: -.5em; '+
    '  top: .4em; '+
    '  width: .9em; '+
    '  height: .9em;  '+
    '  line-height: .9em; '+
    '  text-align: center;  '+
    '  background: #fff; '+
    '  font-size: 80%;  '+
    '  cursor: pointer; '+
    '} '+
    '.outlineCSS li:last-child > .drop { '+
    '  margin-left: 1px;  '+
    '} '+
    '.outlineCSS .drop ~ ul { '+
    '  display: none;  '+
    '} '+
    '.outlineCSS .dropM ~ ul { '+
    '  display: block; '+
    '} '+
    '.outlineNode {color : black; text-decoration: none;} '+
    '.blockOutline { position: fixed; height: 95%; '+
    '  width: 200px; overflow-x: scroll; '+
    '  overflow-y: scroll; left: 0%; padding: 0 10px;'+
    '  margin: 0 10px;border:1mm solid orange;margin:5mm; '+
    '  box-shadow:3mm 3mm 3mm gray; '+
    '  -webkit-box-shadow:3mm 3mm 3mm gray;-moz-box-shadow:3mm 3mm 3mm gray; '+
    '  -ms-box-shadow:3mm 3mm 3mm gray;-o-box-shadow:3mm 3mm 3mm gray; '+
    '  border-radius:2mm;-webkit-border-radius:2mm;-moz-border-radius:2mm; '+
    '  -ms-border-radius:2mm;-o-border-radius:2mm;} '+
    '.blockReport {position: relative; left: 250px;} '+
    '.blockTableOutline {'+
    '  display: block; position: fixed;  width: 200px; '+
    '  overflow: scroll; height : 95%;  '+
    '  border-top-color: grey; '+
    '  border-top-style: inset; '+
    '  border-top-width: 2px;'+
    '  border-bottom-color: grey;'+
    '  border-bottom-style: inset;'+
    '  border-bottom-width: 2px;'+
    '  border-right-color: grey;'+
    '  border-right-style: inset;'+
    '  border-right-width: 2px;'+
    '  border-left-color: grey;'+
    '  border-left-style: inset;'+
    '  border-left-width: 2px;'+
    '  padding: 0 10px;'+
    '  margin: 0 5px;'+
    ' } ';

   outlineJS =
   '<script type="text/javascript"> '+
	 '	(function() { '+
	 '	  var ul = document.querySelectorAll(''.outlineCSS > li:not(:only-child) ul, .outlineCSS ul ul''); '+
	 '	  for (var i = 0; i < ul.length; i++) {  '+
	 '		var div = document.createElement(''div'');  '+
	 '		div.className = ''drop''; '+
	 '		div.innerHTML = ''+'';  '+
	 '		ul[i].parentNode.insertBefore(div, ul[i].previousSibling);  '+
	 '		div.onclick = function() {  '+
	 '		  this.innerHTML = (this.innerHTML == ''+'' ? ''-'' : ''+'')  ;'+
	 '		  this.className = (this.className == ''drop'' ? ''drop dropM'' : ''drop'');  '+
	 '		} '+
	 '	  } '+
	 '	})(); '+
   '</script>';

procedure PrepareHTMLOutline(Outline: TfrxCustomOutline;
            Node: TfrxHTMLOutlineNode; ObjNum: Integer);
procedure WriteHTMLOutline(Node: TfrxHTMLOutlineNode; var html :String);
function GetIdOutlineHTML(Top :Extended; Height :Extended;
          var FOutlineTree :TfrxHTMLOutlineNode; PageH : Extended) :String;

{ Utility routines }

function Float2Str(const Value: Extended; const Prec: Integer = DefaultPrec): String;
function frxPoint2Str(DP: TfrxPoint; const Prec: Integer = DefaultPrec): String; overload;
function frxPoint2Str(X, Y: Extended; const Prec: Integer = DefaultPrec): String; overload;

function GetCursor(Cursor: TCursor): string;
function GetColor(Color: TColor): string;
function GetBorderRadius(Curve: Integer): string;
function IsHasSpecialChars(const s: string): Boolean;
function IsHasHTMLTags(const s: string): Boolean;

function StrFindAndReplace(const Source, Dlm: WideString; SFR: array of WideString): WideString;

function SVGPattern(Formatted, XLine, YLine, Turn: boolean; Color: TColor;
                    LineWidth: Extended; Name: string): string;
procedure CalcGlassRect(Orientation: TfrxGlassFillOrientation;
                        AbsTop, AbsLeft: Extended; var x, y, Width, Height: integer);
function SVGLine(Formatted, ZeroBased: boolean; CSS: TfrxCSSList; Line: TfrxLineView): string;
function SVGDasharray(Style: TfrxFrameStyle; LineWidth: Extended): string;
function SVGUniqueID: string;
function SVGEscapeTextAndAttribute(const s: WideString): WideString;
function SVGStartSpace(const s: WideString): WideString;

const
  spStroke = $1;
  spHTML = $2;

function SVGShapePath(Shape: TfrxShapeView; Options: integer = 0): string;

function GraphicToBase64AnsiString(Graphic: TGraphic): AnsiString;

procedure BitmapFill(Bitmap: TBitmap; Color: TColor);

function PiecewiseLinearInterpolation(Rotation: Integer; X, Y: array of Integer): Extended;

function IsInclude(const Options, Param: LongWord): Boolean;
function IsCJK(WS: WideString): Boolean; // Chinese, Japanese, Korean
function IsDevanagari(WS: WideString): Boolean; // Hindustani

function ArraySum(A: TDoubleArray): Double; overload;
function ArraySum(A: TLongWordDinArray): LongWord; overload;
function ArraySum(A: TIntegerDinArray): Integer; overload;

function ArrayAvg(A: TDoubleArray): Double;

procedure AddLigatureless(FontName: String);
procedure DeleteLigatureless(FontName: String);

function ReverseSlash(const S: String): String;
function frxPictureFormatToStr(Format: TfrxPictureFormat): String;
function PictureFormatFromGraphicHelper(GHelper: TfrxCustomGraphicFormatClass): TfrxPictureFormat;
function GetPicInfo(Pic: TGraphic): TfrxPictureInfo;

procedure TernaryRasterOperationNeeds(dwRop: DWORD; out NeedD, NeedP, NeedS: Boolean);
procedure TernaryRasterOperation(dwRop: DWORD; Destination, Pattern, Source: TBitmap;
  out Result: TBitmap);

implementation
end.
