
{******************************************}
{                                          }
{             FastReport VCL               }
{              Tool controls               }
{                                          }
{         Copyright (c) 1998-2021          }
{            by Fast Reports Inc.          }
{                                          }
{******************************************}

/// <summary>
///   This unit contains classes of the base dialog form used for every form in
///   FastReport. This form handles save and load state, HIDPI messages,
///   localization messages.
/// </summary>
unit frxBaseForm;

interface

{$I frx.inc}

uses
{$IFNDEF FPC}Windows, Messages, {$ENDIF}
  SysUtils, Classes, Graphics, Controls, Forms, IniFiles, frxDPIAwareInt
{$IFDEF FPC}
  , LResources, LCLType, LMessages, LCLIntf, LCLProc, LazarusPackageIntf, LazHelper
{$ENDIF}
{$IFDEF Delphi6}
, Variants
{$ENDIF};

type
  /// <summary>
  ///   This enumeration defines action of preference state. It used to
  ///   save/load/restore form state.
  /// </summary>
  TfrxPreferencesAction = (
    /// <summary>
    ///   Loads preference setting for the form.
    /// </summary>
    frPaLoad,
    /// <summary>
    ///   Saves preference setting for the form. <br />
    /// </summary>
    frPaSave,
    /// <summary>
    ///   Reset preference setting for the form to default. <br />
    /// </summary>
    frPaRestore);
  /// <summary>
  ///   Type of the form settings which should be processed by current form.
  /// </summary>
  TfrxPreferencesType = (
    /// <summary>
    ///   Position of the form.
    /// </summary>
    frPtFormPos,
    /// <summary>
    ///   Form size. <br />
    /// </summary>
    frPtFormSize,
    /// <summary>
    ///   Form visibility at start. <br />
    /// </summary>
    frPtFormVisibility,
    /// <summary>
    ///   Dock information of the form.
    /// </summary>
    frPtFormDockInfo,
    /// <summary>
    ///   Dock position of the form.
    /// </summary>
    frPtFormDockPos,
    /// <summary>
    ///   Dock size of the form.
    /// </summary>
    frPtFormDockSize,
    /// <summary>
    ///   Custom settings(used for controls state).
    /// </summary>
    frPtFormCustom);
  /// <summary>
  ///   Set of preference types.
  /// </summary>
  TfrxPreferencesTypes = set of TfrxPreferencesType;
  /// <summary>
  ///   Defines when form should apply saved preference settings.
  /// </summary>
  TfrxPreferencesLoadEvent = (
    /// <summary>
    ///   At form creation.
    /// </summary>
    peAtCreate,
    /// <summary>
    ///   When form shows. Only at first show.
    /// </summary>
    peAtShowOnce,
    /// <summary>
    ///   When form shows(applies every time form shows).
    /// </summary>
    peAtShow);

    TfrxShortcutAction = (saCopy, saPaste, saCut, saSelectAll, saUndo, saRedo);

  /// <summary>
  ///   This class used to pass messages of the form. internal use only.
  /// </summary>
  TfrxMessageObject = class(TObject)
  public
    Msg: Cardinal;
    WParam: WPARAM;
    FormRect: TRect;
  end;

  /// <summary>
  ///   The base class for dialog forms used in FastReport. This class handles
  ///   save and load state, HIDPI messages,localization messages.
  /// </summary>
  TfrxBaseForm = class(TForm)
  private
    FCurrentFormPPI: Integer;
    FShowed: Boolean;
    FSavedPPI: Integer;
    FNeedUpdatePPI: Boolean;
    FUpdatingPPI: TfrxMessageObject;
    FPreferences: TfrxPreferencesTypes;
    FIsPPIChanging: Boolean;
    FPrefIsLoaded: Boolean;
    FPrefIsSaved: Boolean;
    FPrefIsLocked: Boolean;
    procedure WMDpiChanged(var Message: TMessage); message FRX_WM_DPICHANGED;
    function GetCurrentFormPPI: Integer;
  protected
    FPrefEvent: TfrxPreferencesLoadEvent;
    FHostedControls: array of TControl;
    procedure DoShow; override;
    procedure DoClose(var Action: TCloseAction); override;
    procedure ChangeScale(M, D: Integer); overload; override;
{$IFDEF DELPHI24}
    procedure ChangeScale(M, D: Integer; isDpiChange: Boolean); overload; override;
{$ENDIF}
    procedure BeforePPIChange; virtual;
    procedure AfterPPIChange; virtual;
    procedure AfterPPIMessage; virtual;
    procedure DoUpdateFormPPI(aNewPPI: Integer);
    { FormShowBeforeLoad and FormShowAfterLoad to prevent form flikering on load }
    procedure FormShowBeforeLoad; virtual;
    procedure FormShowAfterLoad; virtual;
    procedure LoadFormPreferences(PreferencesStorage: TObject; DefPreferencesStorage: TObject); virtual;
    procedure SaveFormPreferences(PreferencesStorage: TObject; DefPreferencesStorage: TObject); virtual;
    procedure ResetFormPreferences(PreferencesStorage: TObject); virtual;
    function GetFormSectionName: String; virtual;
    function GetAvailablePreferences: TfrxPreferencesTypes; virtual;
    procedure TranslateControlsByTag(AControl: TControl);
  public
    procedure AfterConstruction; override;
    procedure BeforeDestruction; override;
    constructor Create(AOwner: TComponent); override;
    /// <summary>
    ///   This method should return reference to an object used to save
    ///   preference. By default it's a registry or an ini file object. The top
    ///   most window should mage this object. In FastReport tthe report
    ///   designer and the report preview return this object.
    /// </summary>
    function GetPreferencesStorage(aDefault: Boolean): TObject; virtual;
    /// <summary>
    ///   This method used to process preference action(load/save/restore) with
    ///   default type for current form.
    /// </summary>
    /// <param name="aPrefArction">
    ///   Preference action type.
    /// </param>
    procedure ProcessPreferences(aPrefArction: TfrxPreferencesAction); overload;
    /// <summary>
    ///   This method used to process preference action(load/save/restore) with
    ///   non-default type which passed as a parameter.
    /// </summary>
    /// <param name="aPrefArction">
    ///   Preference action type.
    /// </param>
    /// <param name="PrefTyp">
    ///   Preference types.
    /// </param>
    procedure ProcessPreferences(aPrefArction: TfrxPreferencesAction; PrefTyp: TfrxPreferencesTypes); overload;
    /// <summary>
    ///   Loads form settings from storage object with using preference type
    ///   parameter.
    /// </summary>
    /// <param name="PreferencesStorage">
    ///   Reference to a storage object.
    /// </param>
    /// <param name="DefPreferencesStorage">
    ///   Reference to a storage object with default settings.
    /// </param>
    /// <param name="PrefTyp">
    ///   Settings type.
    /// </param>
    procedure LoadFormPrefType(PreferencesStorage: TObject; DefPreferencesStorage: TObject; PrefTyp: TfrxPreferencesTypes);
    /// <summary>
    ///   Called when window receive message for localization. Updates form
    ///   resource. Each form should override this method for custom resources.
    /// </summary>
    procedure UpdateResouces; virtual;
    /// <summary>
    ///   Moves controls from current form to host control. Each form should
    ///   fill the list with control it can move by filling FHostedControls
    ///   array.
    /// </summary>
    /// <param name="Host">
    ///   Host control for form objects.
    /// </param>
    procedure HostControls(Host: TWinControl); virtual;
    /// <summary>
    ///   Returns back all moved by HostControls function controls back to
    ///   owner form and pass modal result to the form.
    /// </summary>
    /// <param name="AModalResult">
    ///   Modal result of an action.
    /// </param>
    procedure UnhostControls(AModalResult: TModalResult); virtual;
    /// <summary>
    ///   Form calls this method when receives messages that current PPI was
    ///   changed. If Form has controls that doesn't support automatic scaling,
    ///   those controls should be corrected in this method.
    /// </summary>
    /// <param name="aNewPPI">
    ///   New PPI value.
    /// </param>
    procedure UpdateFormPPI(aNewPPI: Integer); virtual;
    /// <summary>
    ///   Returns current PPI value of the form.
    /// </summary>
    property CurrentFormPPI: Integer read GetCurrentFormPPI;
    /// <summary>
    ///   Sends PPI message to a form or a control and all child dock forms.
    /// </summary>
    /// <param name="aWinControl">
    ///   Control that should receive message.
    /// </param>
    /// <param name="aNewPPI">
    ///   New PPI value.
    /// </param>
    procedure SendPPIMessage(aWinControl:TWinControl; aNewPPI: Integer);
    /// <summary>
    ///   Saves form settings to storage object with using preference type
    ///   parameter.
    /// </summary>
    /// <param name="PreferencesStorage">
    ///   Reference to a storage object.
    /// </param>
    /// <param name="DefPreferencesStorage">
    ///   Reference to a storage object with default settings.
    /// </param>
    /// <param name="PrefTyp">
    ///   Settings type.
    /// </param>
    procedure SaveFormPrefType(PreferencesStorage: TObject; DefPreferencesStorage: TObject; PrefTyp: TfrxPreferencesTypes);
{$IFDEF DELPHI24}
    procedure ScaleForPPI(NewPPI: Integer); override;
{$ENDIF}
    property IsPPIChanging: Boolean read FIsPPIChanging;
  end;

  /// <summary>
  ///   The base class for the forms which should save position and sized by
  ///   default to setting storage(registry).
  /// </summary>
  TfrxBaseLoadSavePrefForm = class(TfrxBaseForm)
  protected
    function GetAvailablePreferences: TfrxPreferencesTypes; override;
  end;

implementation
end.
