
{******************************************}
{                                          }
{             FastReport VCL               }
{               Map Helpers                }
{                                          }
{         Copyright (c) 1998-2021          }
{            by Fast Reports Inc.          }
{                                          }
{******************************************}

unit frxMapHelpers;

interface

{$I frx.inc}

uses
  frxClass, Classes, Contnrs, frxXML, Controls, Types,
  {$IFNDEF FPC}
  Windows,
  {$ELSE}
  LCLType, LCLIntf, LCLProc,
  {$ENDIF}
  Graphics, Menus, frxAnaliticGeometry;

type
  TLayerType = (ltMapFile, ltApplication, ltInteractive, ltGeodata);
  TShapeType = (stNone, stPoint, stPolyLine, stPolygon,
    stRect, stDiamond, stEllipse, stPicture, stLegend, stTemplate,
    stMultiPoint, stMultiPolyLine, stMultiPolygon);

  TShapeData = class
  private
    FData: TDoublePointMatrix;
    FShapeType: TShapeType;
    FWidestPartBounds: TfrxRect;
    FShapeCenter: TfrxPoint;
    FPicture: TPicture;
    FConstrainProportions: Boolean;
    FFont: TFont;
    FLegendText: TStringList;
    FTemplateName: String;

    function GetLegend(const Name: String): String;

    function GetPoint: TDoublePoint;
    procedure SetPoint(const Value: TDoublePoint);

    function GetMultiLine(iPart, iPoint: Integer): TDoublePoint;
    procedure SetMultiLine(iPart, iPoint: Integer; const Value: TDoublePoint);
    function GetPartCount: Integer;
    procedure SetPartCount(const Value: Integer);
    function GetMultiLineCount(iPart: Integer): Integer;
    procedure SetMultiLineCount(iPart: Integer; const Value: Integer);
    function GetRect: TDoubleRect;
    procedure SetRect(const Value: TDoubleRect);
    function GetMultiPoint(iPoint: Integer): TDoublePoint;
    function GetMultiPointCount: Integer;
    procedure SetMultiPoint(iPoint: Integer; const Value: TDoublePoint);
    procedure SetMultiPointCount(const Value: Integer);
  protected
    FTags: TStringList; // SparialData
  public
    constructor CreateFull(iParts: Integer; AShapeType: TShapeType; ATags: TStrings);
    constructor CreatePoint(ATags: TStrings; X, Y: Double);
    constructor CreateRect(ATags: TStrings; AShapeType: TShapeType; DR: TDoubleRect);
    constructor CreatePoly(AShapeType: TShapeType; ATags: TStrings; iPoints: Integer);
    constructor CreateEmpty(AShapeType: TShapeType; ATags: TStrings);
    destructor Destroy; override;
    procedure AddPart(iPoints: Integer);
    function IsGetColumnList(List: TStrings): Boolean;
    procedure CalcBounds;
    procedure GetPolyPoints(var Points: TDoublePointArray; iPart: Integer);
    function IsClosed: Boolean;
    function IsMercatorSuitable: Boolean;

    procedure ReadStringList(var SL: TStringList; Reader: TReader);
    procedure ReadTags(Reader: TReader);
    procedure ReadData(Reader: TReader);
    procedure ReadPicture(Reader: TReader);
    procedure ReadFont(Reader: TReader);

    procedure WriteStringList(SL: TStringList; Writer: TWriter);
    procedure WriteTags(Writer: TWriter);
    procedure WriteData(Writer: TWriter);
    procedure WritePicture(Writer: TWriter);
    procedure WriteFont(Writer: TWriter);

    property Point: TDoublePoint read GetPoint write SetPoint;
    property Rect: TDoubleRect read GetRect write SetRect;

    property MultiPoint[iPoint: Integer]: TDoublePoint read GetMultiPoint write SetMultiPoint;
    property MultiPointCount: Integer read GetMultiPointCount write SetMultiPointCount;

    property MultiLine[iPart, iPoint: Integer]: TDoublePoint read GetMultiLine write SetMultiLine;
    property PartCount: Integer read GetPartCount write SetPartCount;
    property MultiLineCount[iPart: Integer]: Integer read GetMultiLineCount write SetMultiLineCount;

    property Legend[const Name: String]: String read GetLegend;
    property ShapeType: TShapeType read FShapeType;
    property WidestPartBounds: TfrxRect read FWidestPartBounds;
    property ShapeCenter: TfrxPoint read FShapeCenter;

    property Tags: TStringList read FTags; // SparialData
    property Picture: TPicture read FPicture;
    property ConstrainProportions: Boolean read FConstrainProportions write FConstrainProportions;
    property Font: TFont read FFont;
    property LegendText: TStringList read FLegendText;
    property TemplateName: String read FTemplateName write FTemplateName;
  end;

  TMapToCanvasCoordinateConverter = class
  private
    FMercatorProjection: Boolean;
    FIsHasData: Boolean;

    procedure SetMercatorProjection(const Value: Boolean);
  protected
    FXmin, FYmin, FXmax, FYmax: Extended; // Map
    FXRange, FYRange, FYmaxTransformed: Extended; // speedup
    FWidth, FHeight, FOffsetX, FOffsetY: Extended; // Canvas

    FShapeActive: Boolean;
    FShapeZoom: Extended;
    FShapeOffset, FShapeCenter: TfrxPoint;
    FUseOffset: Boolean;

    function YTransform(Y: Extended): Extended;
    function ConvertMercator(Y: Extended): Extended;
    function MapTransform(P: TfrxPoint): TfrxPoint; overload;
    function MapTransform(X, Y: Extended): TfrxPoint; overload;
    function ShapeTransform(P: TfrxPoint): TfrxPoint;
  public
    procedure Init;
    procedure IncludeRect(LayerRect: TfrxRect);
    function AspectRatio: Extended;
    procedure SetCanvasSize(CanvasSize: TfrxPoint);
    procedure SetOffset(OffsetX, OffsetY: Extended);

    procedure ReadDFM(Stream: TStream);
    procedure WriteDFM(Stream: TStream);
    procedure Read(Reader: TReader);
    procedure Write(Writer: TWriter);

    procedure IgnoreShape;
    procedure InitShape(AOffsetX, AOffsetY, AZoom: Extended; AShapeCenter: TfrxPoint);

    function Transform(X, Y: Extended): TfrxPoint; overload;
    function Transform(DoublePoint: TDoublePoint): TfrxPoint; overload;
    function Transform(frxPoint: TfrxPoint): TfrxPoint; overload;
    function TransformOffset(frxPoint: TfrxPoint): TfrxPoint; overload;
    function TransformRect(R: TfrxRect): TfrxRect; // Flip Top <--> Bottom

    function CanvasToMap(Canvas: TfrxPoint): TfrxPoint;

    property UseOffset: Boolean read FUseOffset write FUseOffset;
    property MercatorProjection: Boolean read FMercatorProjection write SetMercatorProjection;
    property IsHasData: Boolean read FIsHasData;
  end;

  TTaggedElement = class
  protected
    FTags: TStringList;
  public
    constructor Create;
    destructor Destroy; override;
    procedure AddTag(const stName, stValue: String);

    function IsHaveAllTags(LayerTags: TStrings): Boolean;
    function IsHaveAnyTag(LayerTags: TStrings): Boolean;

    property Tags: TStringList read FTags;
  end;

  TValueDlm = (vdUnknown, vdApostrophe, vdQuote);

  TfrxMapXMLReader = class(TfrxXMLReader)
  protected
    FValueDlm: TValueDlm;
    function IsLastSlash(const InSt: String): Boolean;
    function IsFirstSlash(const InSt: String): Boolean;
    procedure ReadValuedItem(var {$IFDEF Delphi12}NameS, ValueS{$ELSE}Name, Value{$ENDIF}, Text: String);
    procedure DefineValueDlm(Text: String);
    procedure ChangeValueDlm(var Text: String);
  public
    constructor Create(Stream: TStream);
    function IsReadMapXMLRootItem(Item: TfrxXMLItem): Boolean;
    function IsReadMapXMLItem(Item: TfrxXMLItem): Boolean;
  end;

  TfrxMapXMLDocument = class(TfrxXMLDocument)
  protected
    FMapXMLStream: TStream;
    FMapXMLStreamReader: TfrxMapXMLReader;
    procedure InitMapXMLInternal(Stream: TStream);
  public
    procedure InitMapXMLFile(const FileName: String);
    procedure InitMapXMLStream(Stream: TStream);
    procedure DoneMapXMLFile;
    function IsReadItem(Item: TfrxXMLItem): Boolean;
  end;

  TfrxSumStringList = class (TStringList)
  private
    function GetSum(i: Integer): Integer;
    procedure SetSum(i: Integer; const Value: Integer);
  public
    procedure AddSum(st: String);
    procedure SortSum;

    property Sum[i: Integer]: Integer read GetSum write SetSum;
  end;

type
  TfrxClippingRect = class
  private
    FR: TfrxRect;
    FActive: Boolean;
  public
    constructor Create;
    procedure Init(Active: Boolean; R: TfrxRect);
    function IsCircleInside(Circle: TCircle): Boolean;
    function IsPolygonCover(PolyPoints: TPointArray): Boolean; overload;
    function IsPolygonCover(PolyPoints: TDoublePointArray): Boolean; overload;
    function IsPolygonCover(ShapeData: TShapeData; iPart: Integer): Boolean; overload;
    function IsPointInside(P: TfrxPoint): Boolean;
    function IsPolyLineInside(PolyPoints: TPointArray): Boolean; overload;
    function IsPolyLineInside(PolyPoints: TDoublePointArray): Boolean; overload;
    function IsPolyLineInside(ShapeData: TShapeData; iPart: Integer): Boolean; overload;
    function IsPolygonInside(PolyPoints: TPointArray): Boolean; overload;
    function IsPolygonInside(PolyPoints: TDoublePointArray): Boolean; overload;
    function IsPolygonInside(ShapeData: TShapeData; iPart: Integer): Boolean; overload;
    function IsSegmentInside(S: TSegment): Boolean;
    function IsRectInside(Rect: TRect): Boolean;
    function IsDiamondInside(Rect: TRect): Boolean;
  end;

procedure Simplify(Points: TDoublePointArray; Accuracy: Extended; var UsedCount: Integer);
function SafeFileName(FileName: String): String;

procedure Log(Text: String); overload;
procedure Log(Strings: TStrings); overload;
procedure Log(Stream: TStream); overload;
function ToHex(b: Byte): string;
procedure LogOn;
procedure LogOff;

procedure Translate(WinControl: TWinControl);
procedure TranslateMenu(Menu: TMenu);

function PlatformFileName(const FileName: string): string;

function IsMercatorLatitude(YMin, YMax: Extended): Boolean;

implementation
end.
