{******************************************}
{                                          }
{             FastReport VCL               }
{            Various routines              }
{                                          }
{         Copyright (c) 1998-2021          }
{            by Fast Reports Inc.          }
{                                          }
{******************************************}

/// <summary>
///   The unit contains additional helper function used by other FastReports
///   units like type conversions, CRC calculation, image draw and others.
/// </summary>
unit frxUtils;

interface

{$I frx.inc}

uses
  SysUtils, {$IFNDEF FPC}Windows, Messages,{$ENDIF}
  Classes, Graphics, Controls, Forms,
  StdCtrls, Menus, ImgList, ActnList, ComCtrls, frxClass, frxXML, frxCollections
{$IFDEF FPC}
  ,LCLType, LCLIntf, LazHelper, DateUtils
{$IFDEF LCLGTK2}, Printers{$ENDIF}
{$ENDIF}
{$IFDEF DELPHI16}
, System.UITypes
{$ENDIF}
;

type
{$IFNDEF FPC}
  /// <summary>
  ///   This class used to override standard behaviour of metafile draw
  ///   function. By default metafile size includes bottom and right bounds, so
  ///   the real size of metafile is increased by 1. Default draw function in
  ///   VCL tries to compensate it and decreases right and bottom bounds by 1
  ///   in draw method. This doesn't work because PlayEnhMetaFile stretches
  ///   output. As a result output may lose 1 pixel in width and that's very
  ///   critical for outputs like barcodes lines.
  /// </summary>
  TfrxMetafile = class(TMetafile)
  protected
    procedure Draw(ACanvas: TCanvas; const Rect: TRect); override;
  end;
{$ELSE}
  TfrxMetafile = class(TMetafile);
{$ENDIF}
  TfrxRectArea = class
  public
    X, Y, X1, Y1: Extended;
    constructor Create(c: TfrxComponent); overload;
    constructor Create(Left, Top, Right, Bottom: Extended); overload;
    function InterceptsX(a: TfrxRectArea): Boolean;
    function InterceptsY(a: TfrxRectArea): Boolean;
    function InterceptX(a: TfrxRectArea): TfrxRectArea;
    function InterceptY(a: TfrxRectArea): TfrxRectArea;
    function Max(x1, x2: Extended): Extended;
    function Min(x1, x2: Extended): Extended;
  end;

procedure frxParseDilimitedText(aList: TStrings; aText: String; aDelimiter: Char);
/// <summary>
///   Searches for a component on owner by name.
/// </summary>
/// <param name="Owner">
///   Owner component.
/// </param>
/// <param name="Name">
///   Name of the searched component.
/// </param>
function frxFindComponent(Owner: TComponent; const Name: String): TComponent;
/// <summary>
///   Fills list with all found component in application with class type
///   specified in ClassRef parameter. <br />
/// </summary>
/// <param name="Owner">
///   Components owner.
/// </param>
/// <param name="ClassRef">
///   Class reference to search for.
/// </param>
/// <param name="List">
///   Out list parameter.
/// </param>
/// <param name="Skip">
///   Component to skip.
/// </param>
procedure frxGetComponents(Owner: TComponent; ClassRef: TClass;
  List: TStrings; Skip: TComponent);
/// <summary>
///   Returns full name of the component includes full sequences of owners
///   delimired by '.' sign.
/// </summary>
/// <param name="Owner">
///   Owner component.
/// </param>
/// <param name="c">
///   Component reference.
/// </param>
function frxGetFullName(Owner: TComponent; c: TComponent): String;
procedure frxSetCommaText(const Text: String; sl: TStrings; Comma: Char = ';');
function frxRemoveQuotes(const s: String): String;
/// <summary>
///   Converts stream bytes to hex string representation.
/// </summary>
/// <param name="Stream">
///   Reference to stream.
/// </param>
/// <returns>
///   Hex string.
/// </returns>
function frxStreamToString(Stream: TStream): String;
/// <summary>
///   Converts hex string representation to stream bytes and writes them to
///   stream.
/// </summary>
/// <param name="s">
///   String with hex representation of bytes.
/// </param>
/// <param name="Stream">
///   Destination stream.
/// </param>
procedure frxStringToStream(const s: String; Stream: TStream);
function frxStringToHex(const Buffer: WideString): String;
function frxHexToString(const Buffer: String): WideString;
function frxDecToBinBytes(d: Longint): string;
function frxStrToFloat(s: String): Extended;

function frxFloatToStr(d: Extended): String;
function frxRect(ALeft, ATop, ARight, ABottom: Extended): TfrxRect;
function frxPoint(X, Y: Extended): TfrxPoint;
function frxGetBrackedVariable(const Str, OpenBracket, CloseBracket: AnsiString;
  var i, j: Integer): AnsiString;
function frxGetBrackedVariableW(const Str, OpenBracket, CloseBracket: WideString;
  var i, j: Integer): WideString;
procedure frxCommonErrorHandler(Report: TfrxReport; const Text: String);
procedure frxErrorMsg(const Text: String);
procedure frxInfoMsg(const Text: String);
function frxConfirmMsg(const Text: String; Buttons: Integer): Integer;
function frxIsValidFloat(const Value: string): Boolean;
procedure frxAssignImages(Bitmap: TBitmap; dx, dy: Integer;
  ImgList1: TImageList; ImgList2: TImageList = nil);
procedure frxDrawIcon(ImgList: TCustomImageList; aBitmap: TBitmap; IconIndex: Integer; BgColor: TColor = clNone);
procedure frxDrawTransparent(Canvas: TCanvas; x, y: Integer; bmp: TBitmap);
procedure frxDrawCheckBox(Canvas: TCanvas; ARect: TRect; aCheckColor: TColor; aFrameColor: TColor; bIsChecked: Boolean);
procedure frxDrawArrow(Canvas: TCanvas; aRect: TRect; aColor: TColor; bVertical: Boolean = True; bMirrored: Boolean = False);
procedure frxDrawArrows(Canvas: TCanvas; aRect: TRect; aColor: TColor; bVertical: Boolean = True; bMirrored: Boolean = False; LineWidth: Integer = 2; LineGap: Integer = 2; Gap: Integer = 3; Count: Integer = 2);
procedure frxParsePageNumbers(const PageNumbers: String; List: TStrings;
  Total: Integer);
function ColorText(Color: TColor) : String;
function FFColorText(Color: TColor) : String;
function HTMLRGBColor(Color: TColor): string;
procedure frxWriteCollection(Collection: TCollection; Writer: TWriter; Owner: TfrxComponent); overload;
procedure frxReadCollection(Collection: TCollection; Reader: TReader; Owner: TfrxComponent); overload;
procedure frxWriteCollection(Collection: TCollection; const Name: String; Item: TfrxXmlItem; Owner: TfrxComponent; aAncestor: TfrxCollection); overload;
procedure frxReadCollection(Collection: TCollection; Item: TfrxXmlItem; Owner: TfrxComponent;  aAncestor: TfrxCollection); overload;
procedure frxWriteProperties(Instance: TPersistent; Writer: TWriter; Owner: TfrxComponent);
procedure frxReadProperties(Instance: TPersistent; Reader: TReader; Owner: TfrxComponent);

function GetAppFileName: String;
function GetAppPath: String;
function GetTemporaryFolder: String;
function GetTempFile: String;
function frxCreateTempFile(const TempDir: String): String;
{$IFNDEF Delphi7}
function frFloat2Str(const Value: Extended; const Prec: Integer = 2): String;
{$ELSE}
function frFloat2Str(const Value: Extended; const Prec: Integer = 2; const Sep: Char = '.'): String;
{$ENDIF}
function frxReverseString(const AText: string): string;
function frxStreamCRC32(Stream: TStream): Cardinal;
function frxUnixPath2WinPath(const Path: string): string;
{$IFNDEF Delphi6}
function DirectoryExists(const Name: string): Boolean;
{$ENDIF}
{$IFNDEF Delphi7}
function PosEx(const SubStr, S: string; Offset: Cardinal = 1): Integer;
{$ENDIF}
{lcl widgetsets don't like too many Repaint calls, so we use
 update for them, Delphi isn't disturbed}
procedure frxUpdateControl(AControl: TWinControl;
  const ARepaint: Boolean = {$IFDEF NONWINFPC}False{$ELSE}True{$ENDIF});
procedure GetDisplayScale(DevHandle: THandle; IsPrinterHandle: Boolean; var aScaleX, aScaleY: Extended);
procedure TransparentFillRect(aDC: HDC; Left, Top, Right, Bottom: Integer;
  Color: TColor);
function IfStr(Flag: Boolean; const sTrue: string; sFalse: string = ''): string;
function IfInt(Flag: Boolean; const iTrue: integer; iFalse: integer = 0): integer;
function IfReal(Flag: Boolean; const rTrue: Extended; rFalse: Extended = 0.0): Extended;
function IfColor(Flag: Boolean; const cTrue: TColor; cFalse: TColor = clNone): TColor;
function GetStr(const Id: string): string;
function frxFloatDiff(const Val1, Val2: Extended): Boolean;

procedure ScaleBitmap(aBitmap: TBitmap; aScreenPPI: Integer);
function frxSetBkMode(C: TCanvas; bkMode : Integer): Integer;
procedure ChangeGlyphColor(Index: Integer; Color: TColor);
function DateTimeToUTC(DateTime: TDateTime): TDateTime;

type
  TByteDinArray = array of Byte;
  TWordDinArray = array of Word;
  TLongWordDinArray = array of LongWord;
  TIntegerDinArray = array of Integer;

const
  PageBGName = '_pagebackground';
function IsPageBG(Obj: TfrxComponent): Boolean;

implementation
end.
