
{******************************************}
{                                          }
{             FastReport VCL               }
{              FR Variables                }
{                                          }
{         Copyright (c) 1998-2021          }
{            by Fast Reports Inc.          }
{                                          }
{******************************************}

unit frxVariables;

interface

{$I frx.inc}

uses
  SysUtils,
  {$IFNDEF FPC}
  Windows, Messages,
  {$ENDIF}
  Classes, Graphics, Controls, Forms, Dialogs,
  frxXML, frxCollections
{$IFDEF Delphi6}
, Variants
{$ENDIF};


type
  /// <summary>
  ///   The TfrxVariable class represents a variable.
  /// </summary>
  TfrxVariable = class(TfrxCollectionItem)
  private
    FName: String;
    FValue: Variant;
    {$IFDEF FPC}
    procedure ReadData(Reader: TReader);
    procedure WriteData(Writer: TWriter);
    {$ENDIF}
  protected
    function GetInheritedName: String; override;
    {$IFDEF FPC}
    procedure DefineProperties(Filer: TFiler); override;
    {$ENDIF}
  public
    constructor Create(ACollection: TCollection); override;
    procedure Assign(Source: TPersistent); override;
    {$IFDEF FPC}
    property Value: Variant read FValue write FValue;
    {$ENDIF}
    function IsUniqueNameStored: Boolean; override;
  published
    /// <summary>
    ///   Name of a variable. must be unique.
    /// </summary>
    property Name: String read FName write FName;
    {$IFNDEF FPC}
    /// <summary>
    ///   Value of a variable.
    /// </summary>
    property Value: Variant read FValue write FValue;
    {$ENDIF}
  end;

  /// <summary>
  ///   The TfrxVariables class represents a list of variables. It contains all
  ///   methods necessary for working with the list. TfrxReport.Variables
  ///   property contains a link to the instance of this class.
  /// </summary>
  TfrxVariables = class(TfrxCollection)
  private
    function GetItems(Index: Integer): TfrxVariable;
    function GetVariable(Index: String): Variant;
    procedure SetVariable(Index: String; const Value: Variant);
  public
    constructor Create;
    /// <summary>
    ///   Adds a variable to the end of the list.
    /// </summary>
    function Add: TfrxVariable;
    /// <summary>
    ///   Adds a variable to the given position of the list.
    /// </summary>
    function Insert(Index: Integer): TfrxVariable;
    /// <summary>
    ///   Returns the index of a variable with the given name.
    /// </summary>
    function IndexOf(const Name: String): Integer;
    /// <summary>
    ///   Adds a variable to the specified category.
    /// </summary>
    procedure AddVariable(const ACategory, AName: String; const AValue: Variant);
    /// <summary>
    ///   Deletes a category and all its variables.
    /// </summary>
    procedure DeleteCategory(const Name: String);
    /// <summary>
    ///   Deletes a variable.
    /// </summary>
    procedure DeleteVariable(const Name: String);
    /// <summary>
    ///   Returns the list of categories.
    /// </summary>
    procedure GetCategoriesList(List: TStrings; ClearList: Boolean = True);
    /// <summary>
    ///   Returns the list of variables in the specified category.
    /// </summary>
    procedure GetVariablesList(const Category: String; List: TStrings);
    /// <summary>
    ///   Loads the list of variables from file.
    /// </summary>
    procedure LoadFromFile(const FileName: String);
    /// <summary>
    ///   Loads the list of variables from stream.
    /// </summary>
    procedure LoadFromStream(Stream: TStream);
    /// <summary>
    ///   Loads the list of variables from XML node.
    /// </summary>
    procedure LoadFromXMLItem(Item: TfrxXMLItem; OldXMLFormat: Boolean = True);
    /// <summary>
    ///   Saves the list of variables to file.
    /// </summary>
    procedure SaveToFile(const FileName: String);
    /// <summary>
    ///   Saves the list of variables to stream.
    /// </summary>
    procedure SaveToStream(Stream: TStream);
    /// <summary>
    ///   Saves the list of variables to XML node.
    /// </summary>
    procedure SaveToXMLItem(Item: TfrxXMLItem);
    /// <summary>
    ///   The list of variables.
    /// </summary>
    property Items[Index: Integer]: TfrxVariable read GetItems;
    /// <summary>
    ///   Values of variables.
    /// </summary>
    property Variables[Index: String]: Variant read GetVariable
      write SetVariable; default;
  end;

  TfrxArray = class(TCollection)
  private
    function GetItems(Index: Integer): TfrxVariable;
    function GetVariable(Index: Variant): Variant;
    procedure SetVariable(Index: Variant; const Value: Variant);
  public
    constructor Create;
    function IndexOf(const Name: Variant): Integer;
    property Items[Index: Integer]: TfrxVariable read GetItems;
    property Variables[Index: Variant]: Variant read GetVariable
      write SetVariable; default;
  end;


implementation
end.
